(*****************************************************************************)
(*									     *)
(*	       Error Messages file Handler for UniPascal Compiler	     *)
(*			 (c) 1989 George Petrov /YGP/			     *)
(*									     *)
(*               .  *)
(*	  ';' comment                                                        *)
(*	  '*'('Compiler' | 'Run Time' | 'I/O errors' | 'Country = ...')      *)
(*	  <ErrorNo>':' <ErrorText>                                           *)
(*									     *)
(*  Usage: ERRORS [TextFile] [MsgFile] [(/m|/c) | (/e|/x)]		     *)
(*									     *)
(*****************************************************************************)
(*									     *)
(*		  Error Messages File structure (version 1.00)		     *)
(*									     *)
(*          32      .   *)
(*           TYPE  .	     *)
(*         32       :  *)
(*									     *)
(*  - YGP:	  4       #3"YGP" (..Pascal string);*)
(*									     *)
(*  - Version:	        . (   100)    *)
(*									     *)
(*  - Country:	          (MIK,IBM,) *)
(*		  ,       .		     *)
(*									     *)
(*  - Compiler,        ,         *)
(*  - RunTime,	             *)
(*  - IOerrors:     .   :	     *)
(*		     ,      *)
(*		        ,     *)
(*		 offset       TxtStart    *)
(*		         . *)
(*									     *)
(*  - TxtStart:        ,         *)
(*		      Pascal string.   *)
(*									     *)
(*****************************************************************************)

(*) {$R-,I-}						  { UniPascal options }
{*) {$A-,B-,D+,E+,F-,I-,L+,N-,O-,R-,S+,V-}	       { Turbo Pascal options }

program Errors(Output);
  (*) {*) uses UniPascal, UniCRT, UPCinlines; {*}
  type requests = (None, Extract, MakeNew);

       FileHeader = record
	 AnyText: packed array [0..31] of char; { 32   }
	 YGP:	  string[3];			     {     YGP }
	 Version: Cardinal;		    {    1.00 .. 100 }
	 Country: Cardinal;	       { what ASCII standard to use (0 - Any) }

	 TblCompiler,			      { Compiler error messages Table }
	 TblRunTime,			       { RunTime error messages Table }
	 TblIOerrors,				   { I/O error messages Table }
	 TxtStart: LongInt;				     { Error messages }

	 UnUsed1,
	 UnUsed2:  LongInt;
       end { FileHeader };

  var request:	   requests;
      fh:	   FileHeader;
      TxtFile:	   text;
      ErrFile:	   file;
      ErrFileName,
      TxtFileName: string[127];
(*)

  function  SkipSP(const Buff; Offs, Limit: Cardinal): Cardinal;
    external;
  function  GetInteger(const Buff; Offs: Cardinal;
		       Limit: integer; var EndPos: Cardinal): LongInt;
    code $5f, $14, $44;
{*)
var message: text;
{*}

  procedure DumpErrors;

    procedure HasIOerror;
    begin
    if IOresult <> 0
      then begin write(message, 'Can''t read error message file!');
	   halt;
	   end;
    end { HasIOerror };

    procedure DumpText(TblOffset: LongInt; ErrName: string);
      const MaxNum = 256;					      { dummy }
      type Errors = array [0..MaxNum] of record
	     ErrorNo: Integer;
	     Offset: Cardinal;
	   end;
      var Nerrors, nn, n: Cardinal;
	  ErrPtr: ^Errors;
	  heap: pointer;
	  s: string;

    begin { DumpText }
    Mark(heap);
    if TblOffset <> 0
      then begin seek(ErrFile, TblOffset);
	   BlockRead(ErrFile, Nerrors, 2); HasIOerror;
	   if MaxAvail div (*) WordSizeOf{*) SizeOf {*} (ErrPtr^[0]) < Nerrors
	     then begin writeln(TxtFile, 'Too many ', ErrName, ' errors!');
		  (*) return {*) exit {*}
		  end;
           if Nerrors <> 0
             then begin
	          (*) GetMemWords(ErrPtr, WordSizeOf(ErrPtr^[0]) * Nerrors);
	          BlockRead(ErrFile,ErrPtr^,2*WordSizeOf(ErrPtr^[0])*Nerrors);
	          {*) GetMem(ErrPtr, SizeOf(ErrPtr^[0]) * Nerrors);
	          BlockRead(ErrFile, ErrPtr^, SizeOf(ErrPtr^[0]) * Nerrors);
	          {*}
                  end;
	   HasIOerror;
	   if Nerrors <> 0
	     then begin writeln(TxtFile, '* ', ErrName); nn:= 0;
		  write(message, 0: 5, ' ', ErrName, ' errors extracted!');
		  for n:= 0 to Nerrors-1 do
		    begin seek(ErrFile, ErrPtr^[n].Offset + fh.TxtStart);
		    HasIOerror;
		    BlockRead(ErrFile, s[0], 1);	   { read string size }
		    HasIOerror;
		    BlockRead(ErrFile, s[1], length(s));
		    HasIOerror;
		    if s <> ''
		      then begin Inc(nn); write(message, #13, nn: 5);
			   writeln(TxtFile, ErrPtr^[n].ErrorNo:5, ': ', s);
			   if IOresult <> 0
			     then begin
				  writeln(message,#10'Can''t write text file');
				  halt;
				  end;
			   end
		    end { for }
		  end { if Nerrors <> 0 }
	   end { if TblOffset <> 0 };
    release(heap);
    end { DumpText };

  begin { DumpErrors }
  (*) BlockRead(ErrFile, fh, WordSizeOf(fh)*2);
  {*) BlockRead(ErrFile, fh, SizeOf(fh)); {*}
  HasIOerror;
  if fh.YGP <> 'YGP'
    then begin write(message, 'Invalid Error Message File!'); halt end;
  if fh.version <> 100
    then begin write(message, 'Invalid version of Error Message File!');
	 halt
	 end;
  DumpText(fh.TblCompiler, 'Compiler');
  DumpText(fh.TblRunTime,  'Run Time');
  DumpText(fh.TblIOerrors, 'I/O errors');
  end { DumpErrors };

  procedure CreateFile;
    const MaxNum = 256;      {    256       }
    type ErrKind = (Compiler, RunTime, IOerrors, Undefined);
    var kind: ErrKind;
	ErrCnt: array [Compiler..IOerrors] of Cardinal;    {    }
	ErrNo: LongInt;
	Offs: Cardinal;
	line: string;
	Errors: array [Compiler..IOerrors, 0..MaxNum-1] of record
	  ErrorNo: Integer;
	  Offset: Cardinal;
	end;

    function ErrMsgLine: boolean;
      var ndx: Integer;
	  EndPos: Cardinal;
	  Cntr: LongInt;
	  s: string;
    begin
    s:= line;
    Ndx:= SkipSP(s, 1, length(s));		 {    }
    Delete(s, 1, Ndx);					 {    }
    Ndx:= SkipSP(s, length(s), -length(s));		  {    }
    Delete(s, 1+length(s)+Ndx, -Ndx);		   {      }

    if (s = '') or (s[1] = ';')
      then begin (*) return(false);  {*)
	   ErrMsgLine:= false; exit; {*}
	   end;

    if s[1] = '*'
      then begin Delete(s, 1, 1);
	   Ndx:= SkipSP(s, 1, length(s));    {     }
	   Delete(s, 1, Ndx);				 {    }
	   if s = 'Compiler'
	     then kind:= Compiler
	     else if s = 'Run Time'
	       then kind:= RunTime
	     else if s = 'I/O errors'
	       then kind:= IOerrors
	     else if pos('Country', s) = 1
	       then if fh.Country <> 0
		      then begin writeln(message);
			   writeln(message, 'Only One Country info allowed');
			   write(message, line);
			   halt(2);
			   end
		      else begin Delete(s, 1, 7);
			   Ndx:= SkipSP(s, 1, length(s));
			   Delete(s, 1, Ndx);
			   Cntr:= GetInteger(s, 1, length(s), EndPos);
			   if (EndPos = 0) or (IOresult <> 0) or (Cntr < 0)
			     then begin writeln(message);
				  writeln(message, 'Invalid Country No:');
				  write(message, line);
				  halt(4)
				  end;
			   fh.Country:= Cntr;
			   (*) return(false);
			   {*) ErrMsgLine:= false; exit; {*}
			   end
	     else begin writeln(message);
		  writeln(message, 'Invalid Error message type:');
		  write(message, line);
		  halt(2);
		  end;

	   writeln(message); write(message, ErrCnt[kind]: 5, ' ');
	   case kind of
	     Compiler: write(message, 'Compiler errors');
	     RunTime:  write(message, 'RunTime errors');
	     IOerrors: write(message, 'I/O errors');
	     end { case };

	   (*) return(false); {*) ErrMsgLine:= false; exit; {*}
	   end { if * };

    if kind = Undefined
      then begin writeln(message);
	   writeln(message, 'Error message type not specified!');
	   halt(3)
	   end;

    ErrNo:= GetInteger(s, 1, length(s), EndPos);

    if (EndPos = 0) or (IOresult <> 0) or (ErrNo < -32767) or (32767 < ErrNo)
      then begin writeln(message);
	   writeln(message, 'Invalid error message No:');
	   write(message, line);
	   halt(4)
	   end;

    Delete(s, 1, EndPos);
    Ndx:= SkipSP(s, 1, length(s));	   {     }
    Delete(s, 1, Ndx);					 {    }
    if (s <> '') and (s[1] = ':') then Delete(s, 1, 1);
    Ndx:= SkipSP(s, 1, length(s));	       {    ':' }
    Delete(s, 1, Ndx);					 {    }

    line:= s;
    ErrMsgLine:= s <> '';           {        }
    end { ErrMsgLine };

    procedure HasIOerror;
    begin
    if IOresult <> 0
      then begin writeln(message);
	   writeln(message, 'Error writing ', ErrFileName); halt(1);
	   end;
    end { HasIOerror };

  begin { CreateFile }
  (*) FillWord(ErrCnt, WordSizeOf(ErrCnt), 0); Offs:= 0;
      FillWord(fh, WordSizeOf(fh), 0);
  {*) FillChar(ErrCnt, SizeOf(ErrCnt), #0); Offs:= 0;
      FillChar(fh, SizeOf(fh), #0);
  {*}
  kind:= Undefined;
  with fh do
    begin AnyText:= 'UniPascal Error Messages.'#$d#$a#$1a'YGP.';
    YGP:= 'YGP'; Version:= 100; Country:= 0;
    end { fh };

  (*) BlockWrite(ErrFile, fh, WordSizeOf(fh) * 2);
  {*) BlockWrite(ErrFile, fh, SizeOf(fh)); {*}
  HasIOerror;
  write(message, 'Reading Error messages');
  while not eof(TxtFile) do
    begin readln(TxtFile, line);
    if IOresult <> 0
      then begin writeln(message, #10'Can''t read text file'); halt; end;
    if ErrMsgLine
      then with Errors[kind, ErrCnt[kind]] do
	     begin ErrorNo:= ErrNo;
	     Offset:= Offs;
	     Inc(ErrCnt[kind]); write(message, #13, ErrCnt[kind]: 5);
	     if (ErrCnt[kind] = MaxNum) or (65535 - length(line) - 1 < Offs)
	       then begin writeln(message);
		    writeln(message, 'Too many or too long Error messages.');
		    halt(7);
		    end;
	     BlockWrite(ErrFile, line, length(line)+1);
	     HasIOerror;
	     Inc(Offs, length(line) + 1);
	     end { with };
    end { while };

  with fh do (*) {$W-} (*)
    begin
    TblCompiler:= FilePos(ErrFile);
    TblRunTime:=  TblCompiler +2 +sizeof(Errors[Compiler,0]) *ErrCnt[Compiler];
    TblIOerrors:= TblRunTime +2 +sizeof(Errors[RunTime, 0]) * ErrCnt[RunTime];
    TxtStart:= sizeof(fh);
    end { with fh & $W- };

  (*) {$W-} (*)
    begin BlockWrite(ErrFile, ErrCnt[Compiler], 2); HasIOerror{?};
    BlockWrite(ErrFile, Errors[Compiler],
	       sizeof(Errors[Compiler,0])*ErrCnt[Compiler]);
    HasIOerror{?};
    BlockWrite(ErrFile, ErrCnt[RunTime], 2); HasIOerror{?};
    BlockWrite(ErrFile, Errors[RunTime],
	       sizeof(Errors[RunTime,0])*ErrCnt[RunTime]);
    HasIOerror{?};
    BlockWrite(ErrFile, ErrCnt[IOerrors], 2); HasIOerror{?};
    BlockWrite(ErrFile, Errors[IOerrors],
	       sizeof(Errors[IOerrors,0])*ErrCnt[IOerrors]);
    HasIOerror{?};
    seek(ErrFile, 0); HasIOerror{?};
    BlockWrite(ErrFile, fh, sizeof(fh)); HasIOerror;
    end { $W- };
  end { CreateFile };

  procedure Initialize;
    var HasErrFile,
	HasTxtFile:  boolean;
	reqStr:      string[1];
	ErrFileName,
	TxtFileName: string;

    procedure HelpUser;
    begin
    writeln('Usage: ERRORS [TextFile] [ErrFile] [(/m|/c) | (/e|/x)]');
    writeln('  /m or /c - Make (Create) new error message file');
    writeln('  /e or /x - EXtract errors from error message file');
    halt;
    end { HelpUser };

    procedure ParseParameters;
      var Pcnt: ShortCard;
	  parm: string;
    begin
    for Pcnt:= 1 to ParamCount do
      begin ParamStr(Pcnt, parm);
      if (parm[1] = '/') or (parm[1] = '-')
	then begin delete(parm, 1, 1);
	     parm[1]:= UpCase(parm[1]);
	     if (parm = 'M') and (request = None)
	       then request:= MakeNew
	       else if ((parm = 'X') or (parm = 'E')) and (request = None)
		 then request:= Extract
	       else if (request <> None)
		 then begin writeln(message, 'Too many options.');
		      HelpUser
		      end
	       else begin writeln(message, 'Invalid parameter ', parm);
		    HelpUser;
		    end;
	     end
	else if HasTxtFile
	       then if HasErrFile
		      then begin writeln(message, 'Too many FileNames!');
			   HelpUser;
			   end
		      else begin ErrFileName:= parm; HasErrFile:= true; end
	       else begin TxtFileName:= parm; HasTxtFile:= true; end;
      end { while Pcnt <> 0 };
    end { ParseParameters };

    procedure AskUser((*) const (*) text: string;
			  var	    s:	  string;
			      InformOnly: boolean);
      var ndx: ShortCard;

    begin
{$ifNdef myrzel}
    if not InformOnly
      then writeln(message, text, 'not specified')
      else writeln(message, text, s);
{$else}
        (   default )
    write(message, text, s);
    if InformOnly
      then begin writeln(message);
	   (*) return {*) exit {*};
	   end;
    for ndx:= 1 to length(s) do write(message, #8);
{$endif}
    end { AskUser };

  begin { Initialize }
  writeln('UniPascal error message file handler [1.02]',
          ' (c) 1989, 1990 Software R&D Lab.'(*){*)+'     IBM PC/XT/AT'{*});

  HasErrFile:= false; HasTxtFile:= false; request:= None; reqStr:= '';

  ParseParameters;

  if not HasTxtFile
    then TxtFileName:= 'UniPasErr.TXT';
  AskUser('Error message text: ', TxtFileName, HasTxtFile);

  if not HasErrFile
    then begin ErrFileName:= 'UniPas.Err';
         HasErrFile:= true;
         end;
  AskUser('Error message file: ', ErrFileName, HasErrFile);

  if request = None
    then begin AskUser('            Action: ', reqStr, false);
	 if reqSTR <> '' then
	   if UpCase(reqStr[1]) in ['C', 'M']
	     then request:= MakeNew
	     else if UpCase(reqStr[1]) in ['E', 'X']
		    then request:= Extract
		    else write(message, 'No option specified!');
	 end
    else if request = Extract
	   then writeln(message, 'Extracting Error Message Text.')
	   else writeln(message, 'Creating new Error Message file.');

  {                    }
  {         ( ).        }
  if HasErrFile and HasTxtFile and (request <> none)
    then if request = MakeNew
	   then begin
		(*) open(ErrFile, ErrFileName);
		    if IOresult <> 0
		      then begin write('Can''t create ',ErrFileName); halt end;
		    reset(TxtFile, TxtFileName);
		    if IOresult <> 0
		      then begin write('Can''t open ', TxtFileName); halt end;
		{*) assign(ErrFile, ErrFileName);
		    assign(TxtFile, TxtFileName);
		    rewrite(ErrFile, 1);
		    if IOresult <> 0
		      then begin write('Can''t create ',ErrFileName); halt end;
		    reset(TxtFile);
		    if IOresult <> 0
		      then begin write('Can''t open ', TxtFileName); halt end;
		{*}
		end
	   else begin
		(*) reset(ErrFile, ErrFileName);
		    if IOresult <> 0
		      then begin write('Can''t open ',ErrFileName); halt end;
		    rewrite(TxtFile, TxtFileName);
		    if IOresult <> 0
		      then begin write('Can''t create ',TxtFileName); halt end;
		{*) assign(ErrFile, ErrFileName);
		    assign(TxtFile, TxtFileName);
		    reset(ErrFile, 1);
		    if IOresult <> 0
		      then begin write('Can''t open ',ErrFileName); halt end;
		    rewrite(TxtFile);
		    if IOresult <> 0
		      then begin write('Can''t create ',TxtFileName); halt end;
		{*}
		end
    else HelpUser;
  end { Initialize };

begin { Errors }
(*){*) AssignCRT(message); rewrite(message); {*}
Initialize;
if request = Extract
  then DumpErrors
  else CreateFile;
close(ErrFile); close(TxtFile);
end. { Errors }